import re
from typing import Iterator


class Query:
    """SQL query"""

    def __init__(self, query_str: str = None, default_table_prefix: str = None) -> None:
        """SQL query

        Args:
            query_str (str, optional): SQL query string. Defaults to None.
            default_table_prefix (str, optional):
                If project or dataset that configured table have are omitted,
                it will be complement this prefix. Defaults to None.
        """
        self.query_str = query_str
        self.default_table_prefix = default_table_prefix

    def parse_upstairs(self) -> Iterator[dict]:
        """Parse a SQL query string and get upstream table attributes

        Yields:
            Iterator[dict]: upstream table attributes
        """
        # Check the query has cte or not
        cte_pattern = r"(?:with|,)\s*(\w+)\s+as\s*"
        ctes = re.findall(cte_pattern, self.query_str, re.IGNORECASE)

        # Check a boundary that main query starts
        boundary_num = 0
        main_pattern = r"\)[;\s]*select" if any(ctes) else r"select"
        main_search_result = re.search(main_pattern, self.query_str, re.IGNORECASE)
        if main_search_result:
            boundary_num = main_search_result.start()

        # Split the query to 'main' and 'cte'
        query_group = {}
        query_group["main"] = self.query_str[boundary_num:].strip()
        query_group["cte"] = self.query_str[:boundary_num].strip()

        table_pattern = r"(?:from|join)\s+([`.\-\w]+)"
        main_tables_with_cte_alias = re.findall(
            table_pattern, query_group["main"], re.IGNORECASE
        )

        # Exclude cte table alias from main tables
        tables = [table for table in main_tables_with_cte_alias if table not in ctes]

        cte_tables = re.findall(table_pattern, query_group["cte"], re.IGNORECASE)
        tables.extend(cte_tables)

        for table in tables:
            line = [
                i for i, line in enumerate(self.query_str.splitlines()) if table in line
            ][0]

            yield {
                "table_name": solve_table_prefix(table, self.default_table_prefix),
                "line": line + 1,
                "line_str": self.query_str.splitlines()[line],
            }


def solve_table_prefix(table: str, default_table_prefix: str) -> str:
    """Solve table name prefix

    Args:
        table (str): Table name
        default_table_prefix (str):
            If project or dataset that configured table have are omitted,
            it will be complement this prefix. Defaults to None.

    Returns:
        str: Solved table name
    """
    solved_name = table
    if table.count(".") <= 1:
        solved_name = default_table_prefix + "." + solved_name
    return solved_name



import enum
import os
import pathlib
import re
from logging import getLogger
from typing import Iterator, Optional

from google.cloud import storage
from jinja2 import BaseLoader, Environment, FileSystemLoader

logger = getLogger(__name__)


class SourceType(enum.Enum):
    """SQL template source type"""

    FS = "fs"
    GCS = "gcs"

    def __str__(self):
        return self.name


class SQLTemplate:
    """SQL template"""

    def __init__(
        self,
        map_config: dict,
        source_type: SourceType,
        file_path: str,
        bucket: Optional[str] = None,
        project: Optional[str] = None,
        default_table_prefix: Optional[str] = None,
    ):
        """SQL template

        Args:
            map_config (dict): Mapping configuration
            source_type (SourceType): Source type
            file_path (str): SQL file path
            bucket (Optional[str], optional):
                Bucket name where SQL file saved.Defaults to None.
            project (Optional[str], optional):
                Project name where SQL file saved.Defaults to None.
            default_table_prefix (Optional[str], optional):
                If project or dataset that configured table have are omitted,
                it will be complement this prefix. Defaults to None.
        """
        self._map_config = map_config
        self.source_type = source_type
        self.file_path = file_path
        self.bucket = bucket
        self.project = project
        self.default_table_prefix = default_table_prefix
        self.uri = self.set_uri()

    def set_uri(self) -> str:
        """Create uri from file path

        Returns:
            str: uri
        """
        uri = ""
        if self.source_type == SourceType.FS:
            p = pathlib.Path(self.file_path)
            uri = str(p.resolve())
        elif self.source_type == SourceType.GCS:
            uri = f"gs://{self.bucket}/{self.file_path}"
        return uri

    def get_mapped_tables(self) -> Iterator[str]:
        """Get mapped tables

        Yields:
            Iterator[str]: Mapped tables
        """
        for mapping in self._map_config.get("mapping"):
            is_suffix = False
            if mapping.get("file_suffix"):
                is_suffix = self.file_path.endswith(mapping.get("file_suffix"))
            if is_suffix or self.uri == mapping.get("uri"):
                for table in mapping.get("tables"):
                    yield table

    def get_param_list(self) -> list:
        """Get a list of jinja parameters as dict

        Returns:
            list: Jinja parameters as dict
        """
        param_list = []
        for table in self.get_mapped_tables():
            param_list.append(table.get("params"))
        return param_list

    def search_mapped_table(self, params: dict) -> str:
        """Search a mapped table name from a mapping configuration with jinja parameters

        Args:
            params (dict): Jinja parameters

        Returns:
            str: Mapped table name
        """
        mapped_table = None
        for table in self.get_mapped_tables():
            if table.get("params") == params:
                mapped_table = table.get("table")
                break
        return mapped_table

    def get_jinja_params(self) -> dict:
        """Get jinja parameters

        Returns:
            dict: Jinja parameters
        """
        template_str = ""
        if self.source_type == SourceType.FS:
            template_str = self.get_template_str_fs()
        elif self.source_type == SourceType.GCS:
            template_str = self.get_template_str_gcs()

        jinja_expressions = "".join(
            re.findall("{{[^}]*}}", template_str, re.IGNORECASE)
        )
        return re.findall("[^{} ]+", jinja_expressions, re.IGNORECASE)

    def get_template_str_fs(self) -> str:
        """Get file string that read from a file in local file system

        Returns:
            str: File string
        """
        template_str = ""
        with open(self.file_path) as f:
            template_str = f.read()
        return template_str

    def get_template_str_gcs(self) -> str:
        """Get file string that read from a file in GCS

        Returns:
            str: File string
        """
        client = storage.Client(credentials=None, project=self.project)
        bucket = client.get_bucket(self.bucket)
        blob = bucket.blob(self.file_path)
        return blob.download_as_bytes().decode("utf-8")

    def render(self, params: dict) -> str:
        """Render SQL query string from a jinja template

        Args:
            params (dict): Jinja paramters

        Returns:
            str: SQL query string
        """
        query_str = ""
        if self.source_type == SourceType.FS:
            query_str = self.render_fs(params)
        elif self.source_type == SourceType.GCS:
            query_str = self.render_gcs(params)
        return query_str

    def render_fs(self, params: dict) -> str:
        """Render SQL query string from a jinja template on local file system

        Args:
            params (dict): Jinja paramters

        Returns:
            str: SQL query string
        """
        env = Environment(loader=FileSystemLoader(os.path.dirname(self.file_path)))
        jinja_template = env.get_template(os.path.basename(self.file_path))
        return jinja_template.render(params=params)

    def render_gcs(self, params: dict) -> str:
        """Render SQL query string from a jinja template on GCS

        Args:
            params (dict): Jinja paramters

        Returns:
            str: SQL query string
        """
        template_str = self.get_template_str_gcs()
        jinja_template = Environment(loader=BaseLoader()).from_string(template_str)
        return jinja_template.render(params=params)


class TemplateSource:
    """SQL template source"""

    def __init__(self, strl_config: dict, map_config: dict) -> None:
        """SQL template source

        Args:
            strl_config (dict): Stairlight configuration
            map_config (dict): Mapping configuration
        """
        self._strl_config = strl_config
        self._map_config = map_config

    def search(self) -> Iterator[SQLTemplate]:
        """Search SQL template files

        Yields:
            Iterator[SQLTemplate]: SQL template file attributes
        """
        for source in self._strl_config.get("include"):
            type = source.get("type")
            if type.casefold() == SourceType.FS.value:
                yield from self.search_fs(source)
            elif type.casefold() == SourceType.GCS.value:
                yield from self.search_gcs(source)

    def search_fs(self, source: dict) -> Iterator[SQLTemplate]:
        """Search SQL template files from local file system

        Args:
            source (dict): Source attributes of SQL template files

        Yields:
            Iterator[SQLTemplate]: SQL template file attributes
        """
        source_type = SourceType.FS
        path_obj = pathlib.Path(source.get("path"))
        for p in path_obj.glob("**/*"):
            if (
                not re.fullmatch(rf'{source.get("regex")}', str(p))
            ) or self.is_excluded(source_type=source_type, file_path=str(p)):
                logger.debug(f"{str(p)} is skipped.")
                continue
            yield SQLTemplate(
                map_config=self._map_config,
                source_type=source_type,
                file_path=str(p),
                default_table_prefix=source.get("default_table_prefix"),
            )

    def search_gcs(self, source: dict) -> Iterator[SQLTemplate]:
        """Search SQL template files from GCS

        Args:
            source (dict): Source attributes of SQL template files

        Yields:
            Iterator[SQLTemplate]: SQL template file attributes
        """
        source_type = SourceType.GCS
        project = source.get("project")
        client = storage.Client(credentials=None, project=project)
        bucket = source.get("bucket")
        blobs = client.list_blobs(bucket)
        for blob in blobs:
            if (
                not re.fullmatch(rf'{source.get("regex")}', blob.name)
            ) or self.is_excluded(source_type=source_type, file_path=blob.name):
                logger.debug(f"{blob.name} is skipped.")
                continue
            yield SQLTemplate(
                map_config=self._map_config,
                source_type=source_type,
                file_path=blob.name,
                project=project,
                bucket=bucket,
                default_table_prefix=source.get("default_table_prefix"),
            )

    def is_excluded(self, source_type: SourceType, file_path: str) -> bool:
        """Check if the specified file is out of scope

        Args:
            source_type (SourceType): SQL template source type
            file_path (str): SQL template file path

        Returns:
            bool: Return True if the specified file is out of scope
        """
        result = False
        exclude_list = self._strl_config.get("exclude")
        if not exclude_list:
            return result
        for exclude in exclude_list:
            if source_type.value == exclude.get("type") and re.search(
                rf'{exclude.get("regex")}', file_path
            ):
                result = True
                break
        return result



from .query import Query
from .template import SourceType, SQLTemplate, TemplateSource


class Map:
    """Manages functions related to dependency map objects"""

    def __init__(self, strl_config: dict, map_config: dict, mapped: dict = {}) -> None:
        """Manages functions related to dependency map objects

        Args:
            strl_config (dict): Stairlight configuration
            map_config (dict): Mapping configuration
            mapped (dict, optional):
                Mapped file attributes when a mapping configuration file loaded.
                Defaults to {}.
        """
        self.mapped = mapped
        self.unmapped = []
        self._template_source = TemplateSource(
            strl_config=strl_config, map_config=map_config
        )

    def collect_undefined(self, sql_template: SQLTemplate) -> None:
        """Create a list of unmapped file attributes

        Args:
            sql_template (SQLTemplate): SQL template
        """
        self.unmapped.append(
            {
                "sql_template": sql_template,
                "params": sql_template.get_jinja_params(),
            }
        )

    def write_blank(self) -> None:
        """Create a list of unmapped file attributes, if no mapping configuration set"""
        for sql_template in self._template_source.search():
            self.collect_undefined(sql_template)

    def write(self) -> None:
        """Write a dependency map"""
        for sql_template in self._template_source.search():
            param_list = sql_template.get_param_list()
            if param_list:
                for params in param_list:
                    self._remap(sql_template=sql_template, params=params)
            else:
                self._remap(sql_template=sql_template)

    def _remap(self, sql_template: SQLTemplate, params: dict = {}) -> None:
        """Remap a dependency map

        Args:
            sql_template (SQLTemplate): SQL template
            params (dict, optional): Jinja parameters. Defaults to {}.
        """
        downstairs = sql_template.search_mapped_table(params=params)

        # Grep jinja template variables to suggest new configurations
        if not downstairs:
            self.collect_undefined(sql_template)
            return

        query_str = sql_template.render(params=params)
        query = Query(
            query_str=query_str,
            default_table_prefix=sql_template.default_table_prefix,
        )

        if downstairs not in self.mapped:
            self.mapped[downstairs] = {}

        for upstairs_attributes in query.parse_upstairs():
            upstairs = upstairs_attributes["table_name"]
            values = {
                "type": sql_template.source_type.value,
                "file": sql_template.file_path,
                "uri": sql_template.uri,
                "line": upstairs_attributes["line"],
                "line_str": upstairs_attributes["line_str"],
            }
            if sql_template.source_type == SourceType.GCS:
                values["bucket"] = sql_template.bucket
            self.mapped[downstairs][upstairs] = values



import enum
import json
import os
from logging import getLogger
from typing import Union

from .config import MAP_CONFIG_PREFIX, STRL_CONFIG_PREFIX, Configurator
from .map import Map

logger = getLogger(__name__)


class ResponseType(enum.Enum):
    """Enum: Execution result type of up/dowm command"""

    TABLE: str = "table"
    FILE: str = "file"

    def __str__(self):
        return self.name


class SearchDirection(enum.Enum):
    """Enum: Search direction"""

    UP = "upstairs"
    DOWN = "downstairs"

    def __str__(self):
        return self.name


class Node:
    """A Node of singly-linked list"""

    def __init__(self, val, next=None):
        self.val = val
        self.next = None


class StairLight:
    """Table dependency detector"""

    def __init__(self, config_dir=".", load_file=None, save_file=None) -> None:
        """Table dependency detector

        Args:
            config_dir (str, optional):
                Stairlight configuration directory. Defaults to ".".
            load_file ([type], optional):
                A file name of loading results if load option set. Defaults to None.
            save_file ([type], optional):
                A file name of saving results if save option set. Defaults to None.
        """
        self.load_file = load_file
        self.save_file = save_file
        self._configurator = Configurator(dir=config_dir)
        self._mapped = {}
        self._unmapped = []
        self._strl_config = self._configurator.read(prefix=STRL_CONFIG_PREFIX)
        if self._strl_config:
            if self.load_file:
                self._load()
            else:
                self._set()
                if self.save_file:
                    self._save()

    @property
    def mapped(self) -> dict:
        return self._mapped

    @property
    def unmapped(self) -> list:
        return self._unmapped

    def has_strl_config(self) -> bool:
        return self._strl_config is not None

    def _set(self) -> None:
        if not self._strl_config:
            logger.warning(f"{STRL_CONFIG_PREFIX}.y(a)ml' is not found.")
            return

        map_config_prefix = MAP_CONFIG_PREFIX
        if "settings" in self._strl_config:
            settings = self._strl_config["settings"]
            if "mapping_prefix" in settings:
                map_config_prefix = settings["mapping_prefix"]
        map_config = self._configurator.read(prefix=map_config_prefix)

        dependency_map = Map(
            strl_config=self._strl_config,
            map_config=map_config,
        )
        if map_config:
            dependency_map.write()
            self._mapped = dependency_map.mapped
        else:
            dependency_map.write_blank()

        self._unmapped = dependency_map.unmapped

    def init(self) -> str:
        return self._configurator.create_stairlight_template()

    def check(self) -> str:
        if self.load_file:
            logger.warning("Load option is used, skip checking.")
            return None
        elif not self._unmapped:
            return None
        return self._configurator.create_mapping_template(unmapped=self._unmapped)

    def _save(self) -> None:
        with open(self.save_file, "w") as f:
            json.dump(self._mapped, f, indent=2)

    def _load(self) -> None:
        if not os.path.exists(self.load_file):
            logger.error(f"{self.load_file} is not found.")
            exit()
        with open(self.load_file) as f:
            self._mapped = json.load(f)

    def up(
        self,
        table_name: str,
        recursive: bool = False,
        verbose: bool = False,
        response_type: str = ResponseType.TABLE.value,
    ):
        return self.search(
            table_name=table_name,
            recursive=recursive,
            verbose=verbose,
            response_type=response_type,
            direction=SearchDirection.UP,
        )

    def down(
        self,
        table_name,
        recursive=False,
        verbose=False,
        response_type=ResponseType.TABLE.value,
    ):
        return self.search(
            table_name=table_name,
            recursive=recursive,
            verbose=verbose,
            response_type=response_type,
            direction=SearchDirection.DOWN,
        )

    def search(
        self,
        table_name: str,
        recursive: bool,
        verbose: bool,
        response_type: str,
        direction: SearchDirection,
    ) -> Union[list, dict]:
        if verbose:
            return self.search_verbose(
                table_name=table_name,
                recursive=recursive,
                direction=direction,
                searched_tables=[],
                head=True,
            )
        if response_type in [type.value for type in ResponseType]:
            return self.search_plain(
                table_name=table_name,
                recursive=recursive,
                response_type=response_type,
                direction=direction,
                searched_tables=[],
                head=True,
            )
        return None

    def search_verbose(
        self,
        table_name: str,
        recursive: bool,
        direction: SearchDirection,
        searched_tables: list,
        head: bool,
    ) -> dict:
        relative_map = self.get_relative_map(table_name, direction)
        response = {table_name: {}}
        if not relative_map:
            return response

        if recursive:
            for next_table_name in relative_map.keys():
                if head:
                    searched_tables = []
                    searched_tables.append(table_name)

                searched_tables.append(next_table_name)

                if is_cyclic(tables=searched_tables):
                    details = {
                        "table_name": table_name,
                        "next_table_name": next_table_name,
                        "searched_tables": searched_tables,
                    }
                    logger.warning(f"Circular reference detected!: {details}")
                    continue

                next_response = self.search_verbose(
                    table_name=next_table_name,
                    direction=direction,
                    recursive=recursive,
                    searched_tables=searched_tables,
                    head=False,
                )

                if not next_response.get(next_table_name):
                    continue
                relative_map[next_table_name] = {
                    **relative_map[next_table_name],
                    **next_response[next_table_name],
                }

        response[table_name][direction.value] = relative_map
        logger.debug(json.dumps(response, indent=2))
        return response

    def search_plain(
        self,
        table_name: str,
        recursive: bool,
        response_type: str,
        direction: SearchDirection,
        searched_tables: list,
        head: bool,
    ) -> list:
        relative_map = self.get_relative_map(table_name, direction)
        response = []
        if not relative_map:
            return response

        for next_table_name in relative_map.keys():
            if recursive:
                if head:
                    searched_tables = []
                    searched_tables.append(table_name)

                searched_tables.append(next_table_name)

                if is_cyclic(tables=searched_tables):
                    details = {
                        "table_name": table_name,
                        "next_table_name": next_table_name,
                        "searched_tables": searched_tables,
                    }
                    logger.info(f"Circular reference detected!: {details}")
                    continue

                next_response = self.search_plain(
                    table_name=next_table_name,
                    direction=direction,
                    recursive=recursive,
                    response_type=response_type,
                    searched_tables=searched_tables,
                    head=False,
                )
                response = response + next_response

            if response_type == ResponseType.TABLE.value:
                response.append(next_table_name)
            elif response_type == ResponseType.FILE.value:
                response.append(relative_map[next_table_name].get("uri"))
            logger.debug(json.dumps(response, indent=2))

        return sorted(list(set(response)))

    def get_relative_map(self, table_name: str, direction: SearchDirection) -> dict:
        """[summary]

        Args:
            table_name (str): Table name
            direction (SearchDirection): Search direction

        Returns:
            dict: Relative map
        """
        relative_map = {}
        if direction == SearchDirection.UP:
            relative_map = self._mapped.get(table_name)
        elif direction == SearchDirection.DOWN:
            for key in [k for k, v in self._mapped.items() if v.get(table_name)]:
                relative_map[key] = self._mapped[key][table_name]
        return relative_map


def is_cyclic(tables: list) -> bool:
    """Floyd's cycle-finding algorithm

    Args:
        tables (list): Detected tables

    Returns:
        bool: Table dependencies are cyclic or not
    """
    nodes = {}
    for table in tables:
        if table not in nodes.keys():
            nodes[table] = Node(table)
    for i, table in enumerate(tables):
        if not nodes[table].next and i < len(tables) - 1:
            nodes[table].next = nodes[tables[i + 1]]
    slow = fast = nodes[tables[0]]
    while fast and fast.next:
        slow = slow.next
        fast = fast.next.next
        if slow == fast:
            return True
    return False



import glob
import logging
import re
from collections import OrderedDict
from datetime import datetime, timezone

import yaml

from .template import SourceType

MAP_CONFIG_PREFIX = "mapping"
STRL_CONFIG_PREFIX = "stairlight"

logger = logging.getLogger()


class Configurator:
    def __init__(self, dir: str) -> None:
        """Configuration class

        Args:
            path (str): Configuration file path
        """
        self.dir = dir

    def read(self, prefix: str) -> dict:
        """Read a configuration file

        Args:
            prefix (str): Configuration file name prefix

        Returns:
            dict: Results from reading configuration file
        """
        config = None
        pattern = f"^{self.dir}/{prefix}.ya?ml$"
        config_file = [
            p
            for p in glob.glob(f"{self.dir}/**", recursive=False)
            if re.fullmatch(pattern, p)
        ]
        if config_file:
            with open(config_file[0]) as file:
                config = yaml.safe_load(file)
        return config

    def create_stairlight_template(self) -> str:
        """Create a Stairlight template file

        Returns:
            str: Created file name
        """
        file_name = f"{self.dir}/{STRL_CONFIG_PREFIX}.yaml"
        with open(file_name, "w") as f:
            yaml.add_representer(OrderedDict, self.represent_odict)
            yaml.dump(self.build_stairlight_template(), f)
        return file_name

    def create_mapping_template(self, unmapped: list) -> str:
        """Create a mapping template file

        Args:
            unmapped (list): Unmapped results

        Returns:
            str: Mapping template file
        """
        now = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
        file_name = f"{self.dir}/{MAP_CONFIG_PREFIX}_{now}.yaml"
        with open(file_name, "w") as f:
            yaml.add_representer(
                data_type=OrderedDict, representer=self.represent_odict
            )
            yaml.dump(self.build_mapping_template(unmapped), f)
        return file_name

    @staticmethod
    def represent_odict(
        dumper: yaml.Dumper, odict: OrderedDict
    ) -> yaml.nodes.MappingNode:
        """Create a OrderedDict object for dumping a YAML file
        in order of OrderedDict"""
        return dumper.represent_mapping(
            tag="tag:yaml.org,2002:map", mapping=odict.items()
        )

    @staticmethod
    def build_stairlight_template() -> OrderedDict:
        """Create a OrderedDict object for file 'stairlight.config'

        Returns:
            OrderedDict: stairlight.config template
        """
        return OrderedDict(
            {
                "include": [
                    OrderedDict(
                        {
                            "type": "fs",
                            "path": None,
                            "regex": None,
                            "default_table_prefix": None,
                        }
                    ),
                    OrderedDict(
                        {
                            "type": "gcs",
                            "project": None,
                            "bucket": None,
                            "regex": None,
                            "default_table_prefix": None,
                        }
                    ),
                ],
                "exclude": [OrderedDict({"type": None, "regex": None})],
                "settings": {"mapping_prefix": MAP_CONFIG_PREFIX},
            }
        )

    @staticmethod
    def build_mapping_template(unmapped: list) -> OrderedDict:
        """Create a OrderedDict for mapping.config

        Args:
            unmapped (list): unmapped settings that Stairlight detects

        Returns:
            OrderedDict: mapping.config template
        """
        template = OrderedDict({"mapping": []})
        for unmapped_file in unmapped:
            sql_template = unmapped_file["sql_template"]
            values = OrderedDict(
                {
                    "file_suffix": sql_template.file_path,
                    "tables": [OrderedDict({"table": None})],
                }
            )

            params = None
            if "params" in unmapped_file:
                undefined_params = unmapped_file.get("params")
                params = OrderedDict({})
                for param in undefined_params:
                    param_str = ".".join(param.split(".")[1:])
                    params[param_str] = None

            if params:
                values["tables"][0]["params"] = params

            if sql_template.source_type in [SourceType.GCS]:
                values["uri"] = sql_template.uri
                values["bucket"] = sql_template.bucket

            template["mapping"].append(values)

        return template



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['stairlight', 'stairlight.cli']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.10.3', 'PyYAML>=5.0', 'google-cloud-storage>=1.28.1,<2.0.0']

entry_points = \
{'console_scripts': ['stairlight = stairlight.cli.main:main']}

setup_kwargs = {
    'name': 'stairlight',
    'version': '0.1.2',
    'description': 'Table-level data lineage tool',
    'long_description': '# Stairlight\n\n[![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT) [![CI](https://github.com/tosh2230/stairlight/actions/workflows/ci.yml/badge.svg)](https://github.com/tosh2230/stairlight/actions/workflows/ci.yml)\n\nStairlight is a table-level data lineage tool, detects table dependencies from \'Transform\' SQL files such as plain SELECT, \'CREATE TABLE AS SELECT\', \'CREATE VIEW AS SELECT\'.\n\n## Installation\n\n```sh\n$ pip install stairlight\n```\n\n## Getting Started\n\nThere are three steps to use.\n\n```sh\n# Step 1: Initialize and set data location settings\n$ stairlight init\n\'./stairlight.yaml\' has created.\nPlease edit it to set your data sources.\n\n# Step 2: Map SQL files and tables\n$ stairlight check\n\'./mapping_yyyyMMddhhmmss.yaml\' has created.\nPlease map undefined tables and parameters, and append to your latest file.\n\n# Step 3: Get a table dependency map\n$ stairlight\n```\n\n## Description\n\n### Input\n\n- SQL files\n- Configuration files (YAML)\n    - stairlight.yaml: SQL file locations and include/exclude files.\n    - mapping.yaml: Mapping SQL files and tables.\n\n### Output\n\n- Dependency map file (JSON)\n\n    <details>\n\n    <summary>Example</summary>\n\n    ```json\n    {\n        "PROJECT_d.DATASET_e.TABLE_f": {\n            "PROJECT_j.DATASET_k.TABLE_l": {\n                "type": "fs",\n                "file": "tests/sql/main/test_e.sql",\n                "uri": "/foo/bar/stairlight/tests/sql/main/test_e.sql",\n                "line": 1,\n                "line_str": "SELECT * FROM PROJECT_j.DATASET_k.TABLE_l WHERE 1 = 1"\n            },\n            "PROJECT_g.DATASET_h.TABLE_i": {\n                "type": "gcs",\n                "file": "sql/test_b/test_b.sql",\n                "uri": "gs://baz/sql/test_b/test_b.sql",\n                "line": 23,\n                "line_str": "    PROJECT_g.DATASET_h.TABLE_i AS b",\n                "bucket": "stairlight"\n            },\n            "PROJECT_C.DATASET_C.TABLE_C": {\n                "type": "gcs",\n                "file": "sql/test_b/test_b.sql",\n                "uri": "gs://baz/sql/test_b/test_b.sql",\n                "line": 6,\n                "line_str": "        PROJECT_C.DATASET_C.TABLE_C",\n                "bucket": "stairlight"\n            },\n            "PROJECT_d.DATASET_d.TABLE_d": {\n                "type": "gcs",\n                "file": "sql/test_b/test_b.sql",\n                "uri": "gs://baz/sql/test_b/test_b.sql",\n                "line": 15,\n                "line_str": "        PROJECT_d.DATASET_d.TABLE_d",\n                "bucket": "stairlight"\n            }\n        },\n        "PROJECT_j.DATASET_k.TABLE_l": {\n            "PROJECT_d.DATASET_e.TABLE_f": {\n                "type": "fs",\n                "file": "tests/sql/main/test_d.sql",\n                "uri": "/foo/bar/stairlight/tests/sql/main/test_d.sql",\n                "line": 1,\n                "line_str": "SELECT * FROM PROJECT_d.DATASET_e.TABLE_f WHERE 1 = 1"\n            }\n        },\n        "PROJECT_d.DATASET_d.TABLE_d": {\n            "PROJECT_e.DATASET_e.TABLE_e": {\n                "type": "fs",\n                "file": "tests/sql/main/test_f.sql",\n                "uri": "/foo/bar/stairlight/tests/sql/main/test_f.sql",\n                "line": 1,\n                "line_str": "SELECT * FROM PROJECT_e.DATASET_e.TABLE_e WHERE 1 = 1"\n            }\n        },\n    }\n    ```\n\n    </details>\n\n## Configuration\n\nConfig files used for unit test in CI can be found [here](https://github.com/tosh2230/stairlight/tree/main/config).\n\n### stairlight.yaml\n\n\'stairlight.yaml\' is for setting up Stairlight itself.\n\nIt is responsible for specifying the destination of SQL files to be read, and for specifying the prefix of mapping files.\n\nSQL files can be read from the following storage.\n\n- Local file system(with python pathlib module)\n- Google Cloud Storage\n\n### mapping.yaml\n\n\'mapping.yaml\' is used to define the correspondence between SQL files and tables.\n\nThe `params` attribute allows you to reflect settings in [jinja](https://jinja.palletsprojects.com/) template variables embedded in SQL files. If multiple settings are applied to a SQL file using jinja template, the file will be read as if there were the same number of files as the number of settings.\n\n## Sub-command and options\n\n```txt\n$ stairlight --help\nusage: stairlight [-h] [-c CONFIG] [-s SAVE | -l LOAD] {init,check,up,down} ...\n\nA table-level data lineage tool, detects table dependencies from \'Transform\' SQL files.\nWithout positional arguments, return a table dependency map as JSON format.\n\npositional arguments:\n  {init,check,up,down}\n    init                create a new Stairlight configuration file.\n    check               create a new configuration file about undefined mappings.\n    up                  return upstairs ( table | SQL file ) list\n    down                return downstairs ( table | SQL file ) list\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set a Stairlight configuration directory.\n  -s SAVE, --save SAVE  save results to a file\n  -l LOAD, --load LOAD  load results from a file\n```\n\n### init\n\n`init` creates a new Stairlight configuration file.\n\n```txt\n$ stairlight init --help\nusage: stairlight init [-h] [-c CONFIG] [-s SAVE | -l LOAD]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set a Stairlight configuration directory.\n```\n\n### check\n\n`check` creates a new configuration file about undefined mappings.\nThe option specification is the same as `init`.\n\n### up\n\n`up` outputs a list of tables or SQL files located upstream from the specified table.\n\n- Recursive option(`-r` or `--recursive`) is set, Stairlight will find tables recursively and output as a list.\n- Verbose option(`-v` or `--verbose`) is set, Stairlight will add detailed information and output it as a dict.\n\n```txt\n$ stairlight up --help\nusage: stairlight up [-h] [-c CONFIG] [-s SAVE | -l LOAD] -t TABLE [-o {table,file}] [-v] [-r]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set a Stairlight configuration directory.\n  -s SAVE, --save SAVE  save results to a file\n  -l LOAD, --load LOAD  load results from a file\n  -t TABLE, --table TABLE\n                        table name that Stairlight searches for, can be specified multiple times.\n  -o {table,file}, --output {table,file}\n                        output type\n  -v, --verbose         return verbose results\n  -r, --recursive       search recursively\n```\n\n### down\n\n`down` outputs a list of tables or SQL files located downstream from the specified table.\nThe option specification is the same as `up`.\n',
    'author': 'tosh2230',
    'author_email': 'rev.to12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<3.10',
}


setup(**setup_kwargs)



import argparse
import json
from typing import Callable, Union

from stairlight import ResponseType, StairLight


def command_init(stairlight: StairLight, args: argparse.Namespace) -> None:
    """Execute init command

    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments
    """
    message = ""
    stairlight_template_file = stairlight.init()
    if stairlight_template_file:
        message = (
            f"'{stairlight_template_file}' has created.\n"
            "Please edit it to set your data sources."
        )
        exit(message)
    exit()


def command_check(stairlight: StairLight, args: argparse.Namespace) -> None:
    """Execute check command

    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments
    """
    mapping_template_file = stairlight.check()
    if mapping_template_file:
        message = (
            f"'{mapping_template_file}' has created.\n"
            "Please map undefined tables and parameters, "
            "and append to your latest file."
        )
        exit(message)
    exit()


def command_up(
    stairlight: StairLight,
    args: argparse.Namespace,
) -> Union[dict, list]:
    """Execute up command


    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments

    Returns:
        Union[dict, list]: Upstairs results
    """
    return execute_up_or_down(stairlight.up, args)


def command_down(
    stairlight: StairLight,
    args: argparse.Namespace,
) -> Union[dict, list]:
    """Execute down command

    Args:
        stairlight (StairLight): Stairlight class
        args (argparse.Namespace): CLI arguments

    Returns:
        Union[dict, list]: Downstairs results
    """
    return execute_up_or_down(stairlight.down, args)


def execute_up_or_down(
    up_or_down: Callable, args: argparse.Namespace
) -> Union[dict, list]:
    """Execute a command, up or down

    Args:
        up_or_down (Callable): Either command_up() or command_down()
        args (argparse.Namespace): CLI arguments

    Returns:
        Union[dict, list]: Results
    """
    results = []
    for table_name in args.table:
        result = up_or_down(
            table_name=table_name,
            recursive=args.recursive,
            verbose=args.verbose,
            response_type=args.output,
        )
        if len(args.table) > 1:
            results.append(result)
        else:
            return result
    return results


def set_config_parser(parser: argparse.ArgumentParser) -> None:
    """Set a '--config' argument

    Args:
        parser (argparse.ArgumentParser): ArgumentParser
    """
    parser.add_argument(
        "-c",
        "--config",
        help="set a Stairlight configuration directory.",
        type=str,
        default=".",
    )


def set_save_load_parser(parser: argparse.ArgumentParser) -> None:
    """Set arguments, '--save' and '--load'

    Args:
        parser (argparse.ArgumentParser): ArgumentParser
    """
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-s",
        "--save",
        help="Save results to a file.",
        type=str,
        default=None,
    )
    group.add_argument(
        "-l",
        "--load",
        help="Load results from a file.",
        type=str,
        default=None,
    )


def set_up_down_parser(parser: argparse.ArgumentParser) -> None:
    """Set arguments used by up and down

    Args:
        parser (argparse.ArgumentParser): ArgumentParser
    """
    parser.add_argument(
        "-t",
        "--table",
        help=(
            "Table name that Stairlight searches for, "
            "can be specified multiple times."
        ),
        required=True,
        action="append",
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Output type",
        type=str,
        choices=[ResponseType.TABLE.value, ResponseType.FILE.value],
        default=ResponseType.TABLE.value,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Return verbose results.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-r",
        "--recursive",
        help="Search recursively",
        action="store_true",
        default=False,
    )


def create_parser() -> argparse.ArgumentParser:
    """Create a argument parser

    Returns:
        argparse.ArgumentParser: ArgumentParser
    """
    description = (
        "A table-level data lineage tool, "
        "detects table dependencies from 'Transform' SQL files. "
        "Without positional arguments, "
        "return a table dependency map as JSON format."
    )
    parser = argparse.ArgumentParser(description=description)
    set_config_parser(parser=parser)
    set_save_load_parser(parser=parser)

    subparsers = parser.add_subparsers()

    # init
    parser_init = subparsers.add_parser(
        "init", help="Create a new Stairlight configuration file."
    )
    parser_init.set_defaults(handler=command_init)
    set_config_parser(parser=parser_init)

    # check
    parser_check = subparsers.add_parser(
        "check", help="Create a new configuration file about undefined mappings."
    )
    parser_check.set_defaults(handler=command_check)
    set_config_parser(parser=parser_check)

    # up
    parser_up = subparsers.add_parser(
        "up", help="Return upstairs ( table | SQL file ) list."
    )
    parser_up.set_defaults(handler=command_up)
    set_config_parser(parser=parser_up)
    set_save_load_parser(parser=parser_up)
    set_up_down_parser(parser=parser_up)

    # down
    parser_down = subparsers.add_parser(
        "down", help="Return downstairs ( table | SQL file ) list."
    )
    parser_down.set_defaults(handler=command_down)
    set_config_parser(parser=parser_down)
    set_save_load_parser(parser=parser_down)
    set_up_down_parser(parser=parser_down)

    return parser


def main() -> None:
    """CLI entrypoint"""
    parser = create_parser()
    args = parser.parse_args()
    stairlight = StairLight(
        config_dir=args.config, load_file=args.load, save_file=args.save
    )

    result = None
    if hasattr(args, "handler"):
        if args.handler == command_init and stairlight.has_strl_config():
            exit(f"'{args.config}/stairlight.y(a)ml' already exists.")
        elif args.handler != command_init and not stairlight.has_strl_config():
            exit(f"'{args.config}/stairlight.y(a)ml' is not found.")
        result = args.handler(stairlight, args)
    else:
        if not stairlight.has_strl_config():
            exit(f"'{args.config}/stairlight.y(a)ml' is not found.")
        result = stairlight.mapped

    if result:
        print(json.dumps(result, indent=2))


if __name__ == "__main__":
    main()
